#!/usr/bin/perl -w
# Gets the cave out of a running moria with gdb, and can optionally
# put it in another running moria.
# Written to deal with an infinite loop created by a deranged map.
# Benjamin Elijah Griffin	21 Feb 2000

use strict;
use vars qw( $pid $pid2 $gdbskip @cave $i $j $file $err @walls %pos );
use Data::Dumper;

select((select(STDERR), $| = 1)[0]); # autoflush 

$err = 0;
$/ = '(gdb)'; # input record seperator to gdb prompt

$pid = shift;

if ( -f $pid) {
  # Read cave data from dump file.

  print "Reading $pid... ";
  do $pid;
  print " ...done\n";
  $gdbskip = 1;
} elsif (!kill 0, $pid) {
  print "Usage: $0 pid [pid]\n";
  print "or     $0 cave-dump pid\n";
  print "Saves the map from a running moria to a pair of text files.\n";
  print "If a second pid is given, the map is copied to that\n";
  print "program's memory. (Wizards only, as monster and object\n";
  print "lists are not copied.)\n";
  print "If a cave-dump and a pid are given, the dump is transfered to\n";
  print "the running moria. Useful for editing a map and putting it back\n";
  print "in the same moria.\n";
  exit(1);
}

$pid2 = shift;

%pos = (
  cptr => 0,    # index into monster_type m_list[MAX_MALLOC (==125)]
  tptr => 1,	# index into   inven_type t_list[MAX_TALLOC (==175)]
  fval => 2,	# floor value
# These are not used in our mapping, but are used for map from file xfers.
  lr   => 3,	# boolean: room perm light    (eg room floor)
  fm   => 4,	# boolean: has hidden object  (eg trap)
  pl   => 5,	# boolean: other perm light   (eg known walls)
  tl   => 6,	# boolean: temp light         (eg lantern)
);

# From constant.h:
##define NULL_WALL	0
##define DARK_FLOOR	1
##define LIGHT_FLOOR	2
##define MAX_CAVE_ROOM	2
##define CORR_FLOOR	3
##define BLOCKED_FLOOR	4 /* a corridor space with cl/st/se door or rubble */
##define MAX_CAVE_FLOOR	4
#
##define MAX_OPEN_SPACE  3
##define MIN_CLOSED_SPACE 4
#
##define TMP1_WALL	8
##define TMP2_WALL	9
#
##define MIN_CAVE_WALL	12
##define GRANITE_WALL	12
##define MAGMA_WALL	13
##define QUARTZ_WALL	14
##define BOUNDARY_WALL	15

@walls = (
#  0    1    2    3    4    5    6    7    8    9   10   11
  '0', '1', '.', '.', ':', '5', '6', '7', 't', 'T', 'A', 'B',
# 12   13   14   15
  '#', '%', '%', '#',
);

if (!$gdbskip) {

$file = "/tmp/cave.gdb.$$";
print "gdb ./moria $pid > $file\n";
open(CAVE,"|gdb ./moria $pid > $file 2>&1") || die "Can't open pipe to gdb: $!\n";
print CAVE "print cur_width\n";  # should be 198
print CAVE "print cur_height\n"; # should be  66
for($i=0; $i<198; $i ++) {
  for($j=0; $j<66; $j ++) {
    print CAVE "print cave[$j][$i]\n";
  }
}
print CAVE "quit\n";
# Detach from running process? y/n
print CAVE "y\n";
close CAVE;

open(DATA, $file) || die "Can't read data file: $!\n";

$_ = <DATA>; # discard intro text
$_ = <DATA>;

if (!/\$1\s+=\s+198\s+/) {
  print "Invalid width (want 198): $_\n";
  $err ++;
}
$_ = <DATA>;
if (!/\$2\s+=\s+66\s+/) {
  print "Invalid height (want 66): $_\n";
  $err ++;
}

for($j=0; $j<66; $j ++) {
  for($i=0; $i<198; $i ++) {
    $_ = <DATA>;
    die "Bad read from data file: eof/empty, at " unless $_;
    my ($cptr,$tptr,$fval,$lr,$fm,$pl,$tl);
    if (($cptr,$tptr,$fval,$lr,$fm,$pl,$tl) =
      /\$\d+\s+=\s+\{
      cptr\s+=\s+(\d+)\s+'(?:[^']+|\\')', \s+
      tptr\s+=\s+(\d+)\s+'(?:[^']+|\\')', \s+
      fval\s+=\s+(\d+)\s+'(?:[^']+|\\')', \s+
      lr\s+=\s+(\d+), \s+
      fm\s+=\s+(\d+), \s+
      pl\s+=\s+(\d+), \s+
      tl\s+=\s+(\d+)\}/x) {
        $cave[$j][$i] = [$cptr,$tptr,$fval,$lr,$fm,$pl,$tl];
      } else {
        warn "Couldn't parse\n$_\nat ";
	$err ++;
      }
  }
  print STDERR chr($j + 33);
}
print STDERR "\n";

close DATA;
unlink $file unless $err;

open(DUMP,">cave-dump.$pid") or die "Can't open cave-dump.$pid: $!\n";
print "Dump output is to cave-dump.$pid\n";
print DUMP Data::Dumper->Dump([\@cave], [qw(cave)]);
close DUMP;

open(OUT,">cave.$pid") or die "Can't open cave.$pid: $!\n";
print "Map output is to cave.$pid\n";

print OUT <<'KEYTEXT';
Key:   # %  walls                internal use: t T  temp walls type 1 and 2
       .    floor                              0    null wall
       :    obscured floor                     1    dark floor
       o    object                       5 6 7 A B  undefined floor types
       m    monster                            ?    some other unknown
       @    player

Objects include doors, stairs, and traps. Objects (when treasure) can appear
in walls. Internal use things should not show up on the map unless image is
taken at very special times.

KEYTEXT

print OUT (" " x 198) . "\n";
for($j=0; $j<66; $j ++) {
  for($i=0; $i<198; $i ++) {
    my $char;
    if(defined($char = $walls[$cave[$j][$i][$pos{fval}]])) {
      my $obj;
      my $mon;
      if (defined($obj = $cave[$j][$i][$pos{tptr}])) {
        $char = 'o' if ($obj);
      }
      if (defined($mon = $cave[$j][$i][$pos{cptr}])) {
        $char = 'm' if ($mon > 1);
        $char = '@' if ($mon == 1);
      }
      print OUT $char;
    } else {
      print OUT '?';
    }
  }
  print OUT "\n";
}
print OUT (" " x 198) . "\n";
print OUT "\n";
print OUT "vim: set nowrap :\n";
close OUT;

} # end if gdbskip

if (!$err and defined($pid2)) {
  # Now, Igor, we transfer the brain from one to the other!
  # Bwaa-haa-haa-haa!

  print "begin map transfer (slow)\ngdb ./moria $pid2 > /dev/null\n";
  open(TARGET,"|gdb ./moria $pid2 > /dev/null 2>&1")
  	|| die "Can't open pipe to gdb: $!\n";

  select((select(TARGET), $| = 1)[0]); # autoflush 

  for($j=0; $j<66; $j ++) {
    my $olderr = $err;

    for($i=0; $i<198; $i ++) {
      my $char;

      if ($pid != $pid2) {

	if(defined($char = $cave[$j][$i][$pos{fval}])) {
	  # do nothing
	} else {
	  $char = 12; # plain rock
	  warn "cave[$j][$i].fval ($pos{fval}) undefined\n";
	  $err ++;
	}
	print TARGET "set cave[$j][$i].fval = $char\n";

	if ($gdbskip) {
	  # We are using a map from a file, so trust the map on this.
	  &printpart('cptr', $j, $i);
	  &printpart('tptr', $j, $i);
	  &printpart('lr', $j, $i);
	  &printpart('fm', $j, $i);
	  &printpart('pl', $j, $i);
	  &printpart('tl', $j, $i);

	} else {
	  # We are copying a map, so reset all other values.
	  print TARGET "set cave[$j][$i].cptr = 0\n";
	  print TARGET "set cave[$j][$i].tptr = 0\n";
	  print TARGET "set cave[$j][$i].lr = 0\n";
	  print TARGET "set cave[$j][$i].fm = 0\n";
	  print TARGET "set cave[$j][$i].pl = 1\n"; # light spot
	  print TARGET "set cave[$j][$i].tl = 0\n";
	}
      } else {
        # Reading and writing the same process. Just turn on the lights.
	  print TARGET "set cave[$j][$i].pl = 1\n"; # light spot
      }
    }

    print STDERR chr($j + 33);
    print STDERR "\nArgh: ".($err - $olderr)." errors occured for that line\n"
    	if ($olderr != $err);
  }
  print STDERR "\n";

  print TARGET "quit\n";
  # Detach from running process? y/n
  print TARGET "y\n";
  close TARGET;

} # end brain transfer


sub printpart ($$$) {
  my $struct = shift;
  my $j      = shift;
  my $i      = shift;
  my $val    = $cave[$j][$i][$pos{$struct}];

  if (defined($val)) {
    print TARGET "set cave[$j][$i].$struct = $val";
  } else {
    warn "cave[$j][$i].$struct ($pos{$struct}) undefined\n";
    $err ++;
  }
}
__END__
